/*
 * Copyright (c) 1996 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifndef MAKEDEPEND
#endif

#include <DemVtk/demdataset.h>
#include <DemVtk/dempointlocator.h>
#include <DemVtk/demtriangle.h>
#include <TopoFace/fgeomobjs.h>
#include <stdio.h>

DemDataSet::DemDataSet() : vtkPolyData() {
}

void DemDataSet::SetPolyData(vtkPolyData* pd) {
  CopyStructure(pd);
  if ( this->Locator ) this->Locator->Delete();
  this->Locator = DemPointLocator::New();
  this->Locator->SetDataSet(this);
  this->PointData = *pd->GetPointData();
  BuildCellIndex();
}

void DemDataSet::BuildCellIndex() {
  float bounds[6];
  GetBounds(bounds);
  float xwid = (bounds[1] - bounds[0])/4.0;
  float ywid = (bounds[3] - bounds[2])/4.0;
  ids = new vtkIdList*[4*4];
  tileboxes = new FBoxObj*[4*4];
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++) {
      ids[i*4+j] = vtkIdList::New();
      tileboxes[i*4+j] = new FBoxObj(bounds[0]+i*xwid,
				     bounds[2]+j*ywid,
				     bounds[0]+(i+1)*xwid,
				     bounds[2]+(j+1)*ywid);
      for (int k = 0; k < GetNumberOfCells(); k++) {
	DemTriangle* cell = (DemTriangle*)GetCell(k);
	float* cx = new float[3];
	float* cy = new float[3];
	float *pt1, *pt2, *pt3;
	pt1 = cell->Points.GetPoint(1);
	pt2 = cell->Points.GetPoint(2);
	pt3 = cell->Points.GetPoint(0);
	cx[0] = pt1[0];
	cx[1] = pt2[0];
	cx[2] = pt3[0];
	cy[0] = pt1[1];
	cy[1] = pt2[1];
	cy[2] = pt3[1];
	FFillPolygonObj tri(cx, cy, 3);
	if (tri.Intersects(*tileboxes[i*4+j]))
	  ids[i*4+j]->InsertNextId(k);
      }
    }
}

vtkCell* DemDataSet::GetCell(int cellId) {
  static DemTriangle dtriangle;

  int i, loc, numPts, *pts;
  vtkCell *cell = NULL;
  unsigned char type;

  if ( !this->Cells ) this->BuildCells();

  type = this->Cells->GetCellType(cellId);
  loc = this->Cells->GetCellLocation(cellId);

  switch (type)
    {
    case VTK_TRIANGLE:
      cell = &dtriangle;
      this->Polys->GetCell(loc,numPts,pts);
      break;
    default:
      return vtkPolyData::GetCell(cellId);
    }
  // make sure there's enough storage; insert does an allocate, set doesn't
  cell->PointIds.InsertId(numPts-1,pts[numPts-1]);
  cell->Points.InsertPoint(numPts-1,this->Points->GetPoint(pts[numPts-1]));
  
  for (i=0; i<numPts-1; i++)
    {
    cell->PointIds.SetId(i,pts[i]);
    cell->Points.SetPoint(i,this->Points->GetPoint(pts[i]));
    }

  return cell;

}

int DemDataSet::FindCell(float x[3], vtkCell *cell, int cellId, float tol2, int& subId,
                          float pcoords[3], float *weights)
{
  int i, j;
  int closestCell = -1;
  int ptId; //, cellId;
  float dist2, minDist2=VTK_LARGE_FLOAT;
  int sId;
  float pc[3], closestPoint[3];
  static vtkIdList cellIds(VTK_CELL_SIZE);
  static float *w=new float[this->GetMaxCellSize()];

  if ( !this->Points ) return -1;

  FPointObj ptobj(x[0], x[1]);
  for (int b = 0; b < 4*4; b++) {
    if (tileboxes[b]->Contains(ptobj)) {
      
      for (i = 0; i < ids[b]->GetNumberOfIds(); i++) {
	DemTriangle* cell = (DemTriangle*)GetCell(ids[b]->GetId(i));
	if (cell->EvaluatePosition(x,closestPoint,sId,pc,dist2,w) != -1) {
	  return ids[b]->GetId(i);
	}
      }

    }
  }
#if 0
  for (i = 0; i < GetNumberOfCells(); i++) {
    DemTriangle* cell = (DemTriangle*)GetCell(i);
    if (cell->EvaluatePosition(x,closestPoint,sId,pc,dist2,w) != -1) {
      return i;
    }
  }
  //printf("whoa, couldn't find closest point\n"); // these seem to be outside
#endif
#if 0
  if ( !this->Locator )
    {
    this->Locator = new DemPointLocator;
    this->Locator->SetDataSet(this);
    }

  if ( this->Points->GetMTime() > this->Locator->GetMTime() )
    {
    this->Locator->SetDataSet(this);
    }

// Find the closest point to the input position.  Then get the cells that 
// use the point.  Then determine if point is in any of the cells.

  if ( (ptId = this->Locator->FindClosestPoint(x)) >= 0 )
    {
    this->GetPointCells(ptId, cellIds);
    for (i=0; i<cellIds.GetNumberOfIds(); i++)
      {
      cellId = cellIds.GetId(i);
      cell = this->GetCell(cellId);
      if ( cell->EvaluatePosition(x,closestPoint,sId,pc,dist2,w) != -1/* &&
      dist2 <= tol2 && dist2 < minDist2*/ )
        {
        minDist2 = dist2;
        closestCell = cellId;
        subId = sId;
        pcoords[0] = pc[0]; pcoords[1] = pc[1]; pcoords[2] = pc[2]; 
        for (j=0; j < cell->GetNumberOfPoints(); j++) weights[j] = w[j];
	return closestCell;
        }
      }
    printf("none of the closest points' cells seems to contain the point\n");
    if (closestCell == -1) { // try other triangle's points
      this->GetPointCells(ptId, cellIds);
      for (i=0; i<cellIds.GetNumberOfIds(); i++)
	{
	  cellId = cellIds.GetId(i);
	  cell = this->GetCell(cellId);
	  for (int pidi = 0; pidi < cell->PointIds.GetNumberOfIds(); pidi++) {
	    int pid = cell->PointIds.GetId(pidi);
	    vtkIdList pcellIds(VTK_CELL_SIZE);
	    this->GetPointCells(pid, pcellIds);
	    for (j = 0; j<pcellIds.GetNumberOfIds(); j++) {
	      int pcellId = pcellIds.GetId(j);
	      cell = this->GetCell(pcellId);
	      if ( cell->EvaluatePosition(x,closestPoint,sId,pc,dist2,w) != -1 &&
		   dist2 <= tol2 && dist2 < minDist2 )
		{
		  minDist2 = dist2;
		  closestCell = pcellId;
		  subId = sId;
		  pcoords[0] = pc[0]; pcoords[1] = pc[1]; pcoords[2] = pc[2]; 
		  for (int k=0; k < cell->GetNumberOfPoints(); k++) weights[k] = w[k];
		  return closestCell;
		}
	      
	    }
	  }
	}
      printf("didn't find a cell on second attempt either\n");
    }
    }
  else
    printf("whoa, couldn't find closest point\n");
#endif
  return closestCell;
}
