/*
 * Copyright (c) 1997, 1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef mapfunc_h
#define mapfunc_h

#include <ComTerp/comfunc.h>
#include <ComTerp/comterp.h>
#include <OverlayUnidraw/clippoly.h>
#include <ComUnidraw/unifunc.h>

#define VPF_FREE
#define DLG_FREE

class ComTerp;
class MapFeature;
class MapFeatureComp;

class MapFunc : public ComFunc {
public:
    MapFunc(ComTerp*);
    virtual void execute() = 0;
protected:
    void insert_array(MapFeature**& array, int& nf, int& curnf, MapFeature* f);
  MapFeature* copy_point(MapFeature*, MapFeatureComp*);
  MapFeature* copy_line(MapFeature*, MapFeatureComp*);
  MapFeature* copy_area(MapFeature*, MapFeatureComp*);
  MapFeature* copy_junction(MapFeature*, MapFeatureComp*);
};

#ifndef VPF_FREE
class LoadVpfFCFunc : public ComFunc {
public:
    LoadVpfFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "fclass=%s(dbpathstr libnamestr covnamestr fcnamestr :tileid tileidnum) -- import a VPF format feature class"; }
protected:
    int _tileid_symid;
};
#endif

#ifndef DLG_FREE
class Load100KDLGFCFunc : public ComFunc {
public:
    Load100KDLGFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "fclass=%s(dbpathstr libnamestr covnamestr fcnamestr) -- import a 100K DLG format feature class"; }
};
#endif

#define POINT_FEATURE 1
#define LINE_FEATURE 2
#define AREA_FEATURE 3
#define NODE_FEATURE 4
#define FEATURE_CLASS 5

class FeatureFunc : public ComFunc {
public:
  FeatureFunc(ComTerp*);
protected:
  int _id_symid;
};

class PointFeatureFunc : public FeatureFunc {
public:
    PointFeatureFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "feature=%s(x0 y0 :id idnum) -- create point feature"; }
};

class LineFeatureFunc : public FeatureFunc {
public:
    LineFeatureFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "feature=%s(x0 y0 x1 y1 [...] :id idnum) -- create line feature"; }
};

class AreaFeatureFunc : public FeatureFunc {
public:
    AreaFeatureFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "feature=%s(x0 y0 x1 y1 [...] :id idnum) -- create area feature"; }
};

class NodeFeatureFunc : public FeatureFunc {
public:
    NodeFeatureFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "feature=%s(x0 y0 :id idnum) -- create node feature"; }
};

class PolygonAreaFunc : public ComFunc {
public:
    PolygonAreaFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "fltnum=%s(areafeature) -- compute area of areafeature"; }
};

class FeatureClassFunc : public ComFunc {
public:
  FeatureClassFunc(ComTerp*);
  virtual void execute();
    virtual const char* docstring() { 
      return "fclass=%s() -- create a feature class"; }
};

class AddFeatureFunc  : public ComFunc {
public:
  AddFeatureFunc(ComTerp*);
  virtual void execute();
    virtual const char* docstring() { 
      return "fclass=%s(feature fclass) -- add a feature to a feature class"; }
};

class PointDistanceFunc : public ComFunc {
public:
  PointDistanceFunc(ComTerp*);
  virtual void execute();
    virtual const char* docstring() { 
      return "fltnum=%s(pointfeature1 pointfeature2) -- compute distance between point features"; }
};

class NearestPointInLineFunc : public ComFunc {
public:
  NearestPointInLineFunc(ComTerp*);
  virtual void execute();
    virtual const char* docstring() { 
      return "newpointfeature=%s(pointfeature linefeature) -- return the point on a line that is closest to another point"; }
};

class ConvexHullFunc : public ComFunc {
public:
  ConvexHullFunc(ComTerp*);
  virtual void execute();
  virtual const char* docstring() { 
    return "areafeature=%s([feature ...] [fclass ...]) -- return convex hull of feature list as area feature"; }
};

class ShapeSelFunc : public MapFunc {
public:
  ShapeSelFunc(ComTerp*);
  virtual void execute();
  virtual const char* docstring() { 
    return "fclass=%s(fclass shapefeature) -- copy features that spatially intersect a feature's shape"; }
protected:
  boolean node_in_node(MapFeature*, MapFeature*);
  boolean node_in_point(MapFeature*, MapFeature*);
  boolean node_in_line(MapFeature*, MapFeature*);
  boolean node_in_area(MapFeature*, MapFeature*);
  boolean point_in_point(MapFeature*, MapFeature*);
  boolean point_in_line(MapFeature*, MapFeature*);
  boolean point_in_area(MapFeature*, MapFeature*);
  boolean line_in_line(MapFeature*, MapFeature*);
  boolean line_in_area(MapFeature*, MapFeature*);
  boolean area_in_area(MapFeature*, MapFeature*);
};

#if defined(CLIPPOLY)
class ClipRectFunc : public MapFunc {
public:
    ClipRectFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
      return "fclass=%s(fclass x0_int y0_int x1_int y1_int) -- clip with rectangle"; }
};

class ClipPolyFunc : public MapFunc {
public:
    virtual void execute();
protected:
    ClipPolyFunc(ComTerp*);
    ClipOperation _op;
};

class ClipPolyAminBFunc : public ClipPolyFunc {
public:
    ClipPolyAminBFunc(ComTerp*);
    virtual const char* docstring() { 
      return "fclass=%s(fclass areafeature_A) -- clip with polygon feature A minus B"; }
};

class ClipPolyBminAFunc : public ClipPolyFunc {
public:
    ClipPolyBminAFunc(ComTerp*);
    virtual const char* docstring() { 
      return "fclass=%s(fclass areafeature_B) -- clip with polygon feature B minus A"; }
};

class ClipPolyAandBFunc : public ClipPolyFunc {
public:
    ClipPolyAandBFunc(ComTerp*);
    virtual const char* docstring() { 
      return "fclass=%s(fclass areafeature_A) -- clip with polygon feature A and B"; }
};
#endif /* defined(CLIPPOLY) */

class FclassExtentFunc : public ComFunc {
public:
    FclassExtentFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() {
        return "%s(fclass :swe :swn :nee :nen) -- return feature class extent"; }

protected:
    int _swe_symid;
    int _swn_symid;
    int _nee_symid;
    int _nen_symid;
};

class AttrSelFunc : public MapFunc {
public:
    AttrSelFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
        return "fclass=%s(fclass expr_str) -- copy features where exprstr is true"; }
};

class LineEvalFunc : public MapFunc {
public:
    LineEvalFunc(ComTerp*);
    virtual void execute();
};

class ExportFCFunc : public ComFunc {
public:
    ExportFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
	return "%s(fclass [host_str port_int] :file path :paint :socket :bitmap) -- remote\nexport in mapserv (or other) format "; }


protected:
    int _paint_symid;
    int _file_symid;
    int _tileid_symid;
    int _sock_symid;
    int _bitmap_symid;
    int _string_symid;
};

class PrintFCFunc : public ExportFCFunc {
public:
    PrintFCFunc(ComTerp*);
    virtual void execute();
};

#if 0
class PrintFCBackFunc : public ComFunc {
public:
    PrintFCBackFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
        return "%s(fclass :paint :file path) -- print in native mapserv format, \nwith	optional painting and/or saving to a file"; }
};
#endif

class ExportFCGrFunc : public ComFunc {
public:
    ExportFCGrFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
        return "%s(fclass host_str port_int) -- remote export in drawtool format"; }

protected:
    int _gs_str_id;
    int _tr_str_id;
    int _gs_symid;

    friend class ExportFCGrGsFunc;
    friend class ExportFCGrTrFunc;
};

class ExportFCGrGsFunc : public ComFunc {
public:
    ExportFCGrGsFunc(ExportFCGrFunc*);
    virtual void execute();
    virtual const char* docstring() { 
        return "%s(gs_str) -- default graphic state string for features"; }

protected:
    ExportFCGrFunc* _func;
};

class ExportFCGrTrFunc : public ComFunc {
public:
    ExportFCGrTrFunc(ExportFCGrFunc*);
    virtual void execute();
    virtual const char* docstring() { 
        return "%s(tr_str) -- default transformer string for feature classes"; }

protected:
    ExportFCGrFunc* _func;
};

class PrintFCGrFunc : public ExportFCGrFunc {
public:
    PrintFCGrFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
        return "%s(fclass) -- print drawtool format to stdout"; }

};

class ComTerpServ;

class PaintFCFunc : public ComFunc {
public:
    PaintFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
	return "%s(fclass) -- paint features based on gs expressions in painting table"; }

protected:
    int _gs_symid;
};

class PaintFCTblFunc : public ComFunc {
public:
    PaintFCTblFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
	return "painttbl(exprstr gsstr) -- add expression/gs pair to painting table\npainttbl() -- dump painting table to stdout\npainttbl(:reset) -- reset painting table"; }

    static void init_tables();
    static void grow_tables(int gsid, postfix_token* expr, int exprlen);
    static void reset_tables();
    static int eval_exprs(AttributeList*);
protected:
    
    static int _table_len;
    static int _table_siz;
    static int* _gs_table;
    static postfix_token** _expr_table;
    static int* _exprlen_table;
    static ComTerpServ* _gsterp;

    friend class PaintFCFunc;
    friend class ExportFCFunc;

    int _reset_symid;
};

class ExprFCFunc : public ComFunc {
public:
    ExprFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
	return "%s(fclass attribute exprstr) -- assign exprstr result to features attribute "; }

protected:
    ComTerpServ* _terp;
};

class ImportFCFunc : public ComFunc {
public:
    ImportFCFunc(ComTerp*);
    virtual void execute();
    virtual const char* docstring() { 
	return "fclass=%s(path_str) -- import a mapserv format feature class"; }
};

class NumFeaturesFunc : public ComFunc {
public:
  NumFeaturesFunc(ComTerp*);
  virtual void execute();
  virtual const char* docstring() { 
    return "numint=%s(fclass) -- return number of features in feature class"; }
};

class GetFeatureFunc : public MapFunc {
public:
  GetFeatureFunc(ComTerp*);
  virtual void execute();
  virtual const char* docstring() { 
    return "feature=%s(fclass i) -- return copy of ith feature from feature class"; }
};

class OverlayComp;

class MapPasteFunc : public PasteFunc {
public:
    MapPasteFunc(ComTerp*,Editor*,OverlayCatalog* = nil);
    virtual void execute();
    virtual const char* docstring() { 
	return "%s(fccomp [scale]) -- paste map featureclass component into the viewer"; }

protected:
  void SensitizeAll(OverlayComp*,boolean);
  int _sens_symid;
};

#endif

